wcr_chisq <- function(x, y, id, order = 2, ...) {

  ## Subset to complete cases
  if (any(is.na(x) | is.na(y))){
    ind <- which(is.na(x) | is.na(y))
    x <- x[-ind]
    y <- y[-ind]
    id <- id[-ind]
  }
  ## Convert to numeric
  if(is.factor(x)){x <- as.numeric(x)}
  if(is.factor(y)){y <- as.numeric(y)}
  if(is.character(x)){x <- as.numeric(as.factor(x))}
  if(is.character(y)){y <- as.numeric(as.factor(y))}
  ## Convert to character
  if(!is.character(id)){id <- as.character(id)}

  # all options
  METHOD <- "Clustered Pearson's Chi-square test"

  n <- length(x)
  n.clust <- length(unique(id))
  pval <- numeric()

  uniq.x <- as.numeric(sort(unique(x)))
  uniq.y <- as.numeric(sort(unique(y)))
  df0 <- (length(uniq.x) - 1) * (length(uniq.y) - 1)

  dumx <- outer(x, uniq.x, "==")
  dumy <- outer(y, uniq.y, "==")
  dumx <- as.matrix(dumx[, -ncol(dumx)])
  dumy <- as.matrix(dumy[, -ncol(dumy)])

  # averagaed WCR
  px <- apply(as.matrix(aggregate(data.frame(dumx), by = list(id), mean)[, -1]), 2, mean)
  py <- apply(as.matrix(aggregate(data.frame(dumy), by = list(id), mean)[, -1]), 2, mean)

  u.wcr <- matrix(NA, nr = n.clust, nc = df0)
  h0 <- matrix(NA, nr = n, nc = df0)
  dumx.p <- sweep(dumx, 2, px, "-")
  dumy.p <- sweep(dumy, 2, py, "-")
  k <- 1
  for (j in 1:ncol(dumy)) {
    for (i in 1:ncol(dumx)) {
      h0[, k] <- dumx.p[, i] * dumy.p[, j]
      u.wcr[, k] <- tapply(dumx.p[, i] * dumy.p[, j], id, mean)
      k <- k + 1
    }
  }
  u <- as.matrix(apply(u.wcr, 2, mean))  #4x1
  h0bar <- as.matrix(aggregate(data.frame(h0), by = list(id), mean)[, -1])

  var0 <- t(h0bar) %*% h0bar/n.clust

  ## First Order Approximation of Variance
  if (order == 1) {
    METHOD <- paste(METHOD, "using a First Order Approximation of Variance",
                    sep = " ")
    v0 <- n.clust * t(u) %*% solve(var0) %*% u

    ## First Order Pval under null
    pval <- 1 - pchisq(v0, df0)

    names(v0) <- "chisq"
    results <- list(statistic = v0, p.value = pval, n.clust = n.clust, nobs = n,
                    method = METHOD, alternative = NULL, null.value = NULL)
  }

  ## Second Order Approximation of Variance
  if (order == 2) {
    METHOD <- paste(METHOD, "using a Second Order Approximation of Variance",
                    sep = " ")
    hbar20 <- matrix(NA, nr = choose(n.clust, 2), nc = df0)
    k <- 1
    for (j in 1:ncol(dumy)) {
      for (i in 1:ncol(dumx)) {
        temp <- outer(dumx[, i], dumx[, i], "-") * outer(dumy[, j], dumy[, j], "-")
        temp <- t(aggrMean(t(aggrMean(temp, id)), id))
        hbar20[, k] <- gdata::upperTriangle(temp - matrix(h0bar[, k], n.clust, n.clust, byrow = T) - matrix(h0bar[,
                                                                                                                  k], n.clust, n.clust, byrow = F))
        k <- k + 1
      }
    }

    varh20 <- t(hbar20) %*% hbar20/(n.clust * (n.clust - 1))
    var20 <- var0 + varh20/(2 * (n.clust - 1))
    chi20 <- n.clust * t(u) %*% solve(var20) %*% u

    ## Second Order Pval under null
    pval <- 1 - pchisq(chi20, df0)

    names(chi20) <- "chisq"
    results <- list(statistic = chi20, p.value = pval, n.clust = n.clust, nobs = n,
                    method = METHOD, null.value = NULL,
                    x = x, y = y, id = id)
  }
  class(results) <- "cluscor"
  return(results)
}
